#include <iostream>
#include <vector>
#include <algorithm>
#include <thread>
#include <mutex>

class Observer {
public:
	virtual void Update(int value) = 0;
};

class ConcreteObserver : public Observer {
public:
	ConcreteObserver(int id) : id_(id) {}

	void Update(int value) override {
		std::cout << "Observer " << id_ << " received update: " << value << std::endl;
	}

private:
	int id_;
};

class Subject {
public:
	virtual void RegisterObserver(Observer* observer) = 0;
	virtual void RemoveObserver(Observer* observer) = 0;
	virtual void NotifyObservers() = 0;
};

class ConcreteSubject : public Subject {
public:
	void RegisterObserver(Observer* observer) override {
		std::unique_lock<std::mutex> lock(observers_mutex_);
		observers_.push_back(observer);
	}

	void RemoveObserver(Observer* observer) override {
		std::unique_lock<std::mutex> lock(observers_mutex_);
		auto it = std::find(observers_.begin(), observers_.end(), observer);
		if (it != observers_.end()) {
			observers_.erase(it);
		}
	}

	void NotifyObservers() override {
		std::unique_lock<std::mutex> lock(observers_mutex_);
		for (Observer* observer : observers_) {
			std::thread t(&Observer::Update, observer, value_);
			t.detach();
		}
	}

	void SetValue(int value) {
		value_ = value;
		NotifyObservers();
	}

private:
	std::vector<Observer*> observers_;
	std::mutex observers_mutex_;
	int value_;
};

//int main() {
//	ConcreteSubject subject;
//	ConcreteObserver observer1(1);
//	ConcreteObserver observer2(2);
//
//	subject.RegisterObserver(&observer1);
//	subject.RegisterObserver(&observer2);
//
//	subject.SetValue(123);
//
//	subject.RemoveObserver(&observer1);
//
//	subject.SetValue(456);
//
//	return 0;
//}
